"""Monkey patching."""

import dacite

from collections.abc import Mapping
from dacite.core import Config, _build_value, extract_generic, is_subclass
from typing import Any


def _build_value_for_collection(
    collection: type,
    data: Any,  # noqa: ANN401
    config: Config,
) -> Any:  # noqa: ANN401
    data_type = type(data)
    if isinstance(data, dict) and is_subclass(collection, Mapping):
        types = extract_generic(collection, defaults=(Any, Any))
        return data_type(
            (key, _build_value(type_=types[1], data=value, config=config))
            for key, value in data.items()
        )
    elif isinstance(data, (list, tuple)) and is_subclass(collection, tuple):
        types = extract_generic(collection)
        if Ellipsis in types:
            return data_type(
                _build_value(type_=types[0], data=item, config=config) for item in data
            )
        elif len(data) == len(types):
            return data_type(
                _build_value(type_=type_, data=item, config=config)
                for item, type_ in zip(data, types, strict=True)
            )
    elif isinstance(data, (list, tuple)):
        types = extract_generic(collection, defaults=(Any,))
        return data_type(_build_value(type_=types[0], data=item, config=config) for item in data)
    return data


dacite.core._build_value_for_collection = _build_value_for_collection



"""CLI utils and parsing."""

from __future__ import annotations

__all__ = [
    "ArgumentParser",
    "choice",
]

import argparse
import copy
import json

from collections.abc import Callable
from dataclasses import MISSING, dataclass, field, is_dataclass
from dataclasses import fields as iter_fields
from functools import partial
from types import UnionType
from typing import (
    Generic,
    Literal,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from unittest.mock import patch

from .core import from_dict
from .typing import Dataclass

T = TypeVar("T", bound=type)
DC = TypeVar("DC", bound=Dataclass)
HELP = "<HELP>"


@dataclass
class ChoiceSpec(Generic[DC]):
    options: dict[str, type[DC]]
    default: str | None = None


@dataclass
class DataclassSpec:
    data_cls: type[Dataclass]
    keys: set[str]
    choices: dict[str, ChoiceSpec]
    root: bool


def origin(t: T) -> T:
    return t if get_origin(t) is None else get_origin(t)


def boolean(s: str) -> bool:
    return s.lower() in {"1", "true", "yes", "on"}


def dict_parser(t: T) -> Callable[[str], T]:
    return lambda s: json.loads(s)


def tuple_parser(t: T) -> Callable[[str], T]:
    types = get_args(t)
    types = types if types else (str, Ellipsis)

    if Ellipsis in types:
        parser = any_parser(types[0])
        return lambda s: tuple(parser(x) for x in s.split())
    else:
        parsers = map(any_parser, types)
        return lambda s: tuple(parser(x) for parser, x in zip(parsers, s.split(), strict=True))


def list_parser(t: T) -> Callable[[str], T]:
    types = get_args(t)
    types = types if types else (str,)

    parser = any_parser(types[0])
    return lambda s: [parser(x) for x in s.split()]


def any_parser(t: T) -> Callable[[str], T]:
    if origin(t) in (int, float, str):
        return origin(t)
    if origin(t) is bool:
        return boolean
    elif origin(t) is dict:
        parser = dict_parser(t)
    elif origin(t) is tuple:
        parser = tuple_parser(t)
    elif origin(t) is list:
        parser = list_parser(t)
    else:
        return str

    parser.__name__ = origin(t).__name__

    return parser


def choice(
    options: dict[str, type[DC]],
    default: str | None = None,
    default_factory: Callable[[], DC] | None = None,
) -> DC:
    """Create a dataclass field that gives a choice between named options."""

    for option in options.values():
        assert is_dataclass(option) and not is_dataclass(type(option))

    if default_factory is None and default is not None:
        default_factory = options[default]

    metadata = {"heracls_choice": ChoiceSpec(options, default)}

    if default_factory is None:
        return field(metadata=metadata)
    else:
        return field(default_factory=default_factory, metadata=metadata)


class SimpleHelpFormatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.MetavarTypeHelpFormatter,
):
    def _get_default_metavar_for_optional(self, action: argparse.Action) -> str:
        return getattr(action.type, "__name__", None)

    def _get_help_string(self, *args, **kwargs) -> str | None:  # noqa: ANN002, ANN003
        help = super()._get_help_string(*args, **kwargs)
        if help is not None:
            help = help.replace(HELP, "")
        return help


class ArgumentParser(argparse.ArgumentParser):
    """Simple dataclass-aware argument parser."""

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        kwargs.setdefault(
            "formatter_class",
            partial(
                SimpleHelpFormatter,
                max_help_position=64,
                width=96,
            ),
        )

        super().__init__(*args, **kwargs)

        self.heracls_specs: dict[str, DataclassSpec] = {}

    def add_arguments(
        self,
        data_cls: type[Dataclass],
        dest: str = "config",
        root: bool = False,
    ) -> None:
        assert is_dataclass(data_cls) and not is_dataclass(type(data_cls))
        assert dest not in self.heracls_specs

        spec = DataclassSpec(
            data_cls=data_cls,
            keys=set(),
            choices={},
            root=root,
        )

        assert dest not in self.heracls_specs

        self.heracls_specs[dest] = spec
        self._register_dataclass(spec, data_cls, () if root else (dest,))

    def _register_dataclass(
        self,
        spec: DataclassSpec,
        data_cls: type[Dataclass],
        path: tuple[str, ...] = (),
    ) -> None:
        if path:
            group = self.add_argument_group(".".join(path))
        else:
            group = self.add_argument_group()

        for f in iter_fields(data_cls):
            f_path = (*path, f.name)
            f_key = ".".join(f_path)
            f_flag = f"--{f_key}"

            kwargs = {"dest": f_key, "help": HELP}

            if "heracls_choice" in f.metadata:
                spec.choices[f_key] = f.metadata["heracls_choice"]

                options = f.metadata["heracls_choice"].options
                default = f.metadata["heracls_choice"].default

                if default is None:
                    kwargs["required"] = True
                else:
                    kwargs["default"] = default

                group.add_argument(f_flag, choices=tuple(options.keys()), **kwargs)
            elif is_dataclass(f.type):
                self._register_dataclass(spec, f.type, f_path)
            else:
                spec.keys.add(f_key)

                if f.default is MISSING and f.default_factory is MISSING:
                    kwargs["required"] = True
                elif f.default is MISSING:
                    kwargs["default"] = f.default_factory()
                else:
                    kwargs["default"] = f.default

                f_type = f.type

                if origin(f_type) == Union or isinstance(f_type, UnionType):
                    types = set(get_args(f_type))
                    types.remove(type(None))
                    if len(types) == 1:
                        f_type = types.pop()

                if origin(f_type) == Literal:
                    types = get_args(f_type)
                    group.add_argument(f_flag, choices=types, **kwargs)
                elif origin(f_type) is tuple:
                    types = get_args(f_type)
                    types = types if types else (str, Ellipsis)

                    if Ellipsis in types:
                        group.add_argument(f_flag, nargs="+", type=any_parser(types[0]), **kwargs)
                    elif all(t == types[0] for t in types):
                        group.add_argument(
                            f_flag, nargs=len(types), type=any_parser(types[0]), **kwargs
                        )
                    else:
                        group.add_argument(f_flag, type=any_parser(f_type), **kwargs)
                elif origin(f_type) is list:
                    types = get_args(f_type)
                    types = types if types else (str,)

                    group.add_argument(f_flag, nargs="*", type=any_parser(types[0]), **kwargs)
                elif origin(f_type) is dict:
                    group.add_argument(f_flag, type=any_parser(f_type), **kwargs)
                else:
                    group.add_argument(f_flag, type=any_parser(f_type), **kwargs)

    def _finalize(
        self,
        args: list[str] | None = None,
        namespace: argparse.Namespace | None = None,
    ) -> ArgumentParser:
        clone = copy.deepcopy(self)
        visited = set()

        while True:
            with (
                patch("argparse.ArgumentParser.error"),
                patch("argparse._HelpAction.__call__"),
            ):
                namespace, _ = argparse.ArgumentParser.parse_known_args(
                    clone, args=args, namespace=namespace
                )

            finished = True
            for spec in clone.heracls_specs.values():
                for key, choice in spec.choices.items():
                    if key not in visited:
                        visited.add(key)
                        value = getattr(namespace, key, None)
                        if value is not None:
                            clone._register_dataclass(spec, choice.options[value], key.split("."))
                            finished = False

            if finished:
                break

        return clone

    def parse_known_args(
        self,
        args: list[str] | None = None,
        namespace: argparse.Namespace | None = None,
    ) -> tuple[argparse.Namespace, list[str]]:
        clone = self._finalize(args, namespace)

        namespace, unknown = argparse.ArgumentParser.parse_known_args(
            clone, args=args, namespace=namespace
        )

        dest_map = {}
        choices = set()

        for dest, spec in clone.heracls_specs.items():
            for key in spec.keys:
                dest_map[key] = dest
            choices.update(spec.choices)

        data_dicts = {dest: {} for dest in clone.heracls_specs}
        others = {}

        for key, value in vars(namespace).items():
            if key in dest_map:
                data_dicts[dest_map[key]][key] = value
            elif key in choices:
                pass
            else:
                others[key] = value

        for dest, spec in clone.heracls_specs.items():
            data = {}
            for key, value in data_dicts[dest].items():
                temp = data
                *keys, leaf = key.split(".")
                for k in keys:
                    temp = temp.setdefault(k, {})
                temp[leaf] = value

            others[dest] = from_dict(spec.data_cls, data)

        namespace = argparse.Namespace(**others)

        return namespace, unknown



"""Tests for the heracls module."""

from dataclasses import dataclass, field
from typing import Literal

from heracls import ArgumentParser, choice, from_yaml, to_yaml


def test_pipeline() -> None:
    @dataclass
    class ModelConfig:
        name: str = "mlp"
        depth: int = 3
        width: int = 256
        dropout: float | None = None

    @dataclass
    class AdamConfig:
        name: Literal["adam"] = "adam"
        betas: tuple[float, float] = (0.95, 0.95)
        learning_rate: float = 1e-3
        weight_decay: float = 0.0

    @dataclass
    class SGDConfig:
        name: Literal["sgd"] = "sgd"
        momentum: float = 0.0
        learning_rate: float = 1e-3
        weight_decay: float = 0.0
        nesterov: bool = False

    @dataclass
    class TrainConfig:
        model: ModelConfig = field(default_factory=ModelConfig)
        optimizer: AdamConfig | SGDConfig = choice(
            {"adam": AdamConfig, "sgd": SGDConfig},
            default="adam",
        )
        dataset: str = "mnist"
        data_splits: tuple[float, ...] = (0.8, 0.1)
        n_epochs: int = 1024
        n_steps_per_epoch: int = 256
        tasks: list[str] = field(default_factory=list)
        magic: dict[str, float] = field(default_factory=dict)

    parser = ArgumentParser()
    parser.add_argument("--dry", action="store_true", help="dry run")
    parser.add_arguments(TrainConfig, dest="train", root=True)

    parser.print_help()

    args = parser.parse_args([
        "--model.dropout",
        "0.1",
        "--optimizer",
        "sgd",
        "--data_splits",
        "0.7",
        "0.2",
        "--magic",
        r'{"p": 3, "i": 0.14}',
    ])

    assert args.train == TrainConfig(
        model=ModelConfig(dropout=0.1),
        optimizer=SGDConfig(),
        data_splits=(0.7, 0.2),
        magic={"p": 3, "i": 0.14},
    )

    dump = to_yaml(args.train)

    assert from_yaml(TrainConfig, dump) == args.train



"""Core data transformations."""

__all__ = [
    "from_dict",
    "from_dotlist",
    "from_omega",
    "from_yaml",
    "to_dict",
    "to_omega",
    "to_yaml",
]

import dacite
import yaml

from dataclasses import asdict, is_dataclass
from dataclasses import fields as iter_fields
from omegaconf import DictConfig, OmegaConf
from typing import Any, TypeVar

from .typing import Dataclass

DC = TypeVar("DC", bound=Dataclass)


def from_dict(data_cls: type[DC], data: dict[str, Any]) -> DC:
    """Instantiate a dataclass from a dictionary.

    Arguments:
        data_cls: A dataclass type.
        data: A dictionary.

    Returns:
        A `data_cls` instance.
    """
    return dacite.from_dict(
        data_class=data_cls,
        data=data,
        config=dacite.Config(
            cast=[tuple],
            strict=True,
            strict_unions_match=True,
        ),
    )


def to_dict(data: Dataclass, *, recursive: bool = True) -> dict[str, Any]:
    """Convert a dataclass instance to a dictionary.

    Arguments:
        data: A dataclass instance.
        recursive: Recurse into nested dataclasses, dicts, lists, and tuples when `True`.

    Returns:
        A dictionary representation of `data`.
    """
    if not is_dataclass(type(data)):
        return data
    elif recursive:
        return asdict(data)
    keys = [f.name for f in iter_fields(data)]
    data = {k: getattr(data, k) for k in keys}
    return data


def from_omega(data_cls: type[DC], data: DictConfig) -> DC:
    """Instantiate a dataclass from an :mod:`omegaconf` config.

    Arguments:
        data_cls: A dataclass type.
        data: A config object.

    Returns:
        A `data_cls` instance.
    """
    return from_dict(
        data_cls,
        OmegaConf.to_container(data, resolve=True, throw_on_missing=True),
    )


def to_omega(data: Dataclass) -> DictConfig:
    """Convert a dataclass instance to an :mod:`omegaconf` config.

    Arguments:
        data: A dataclass instance.

    Returns:
        A config representation of `data`.
    """
    return OmegaConf.create(to_dict(data, recursive=True))


def from_yaml(data_cls: type[DC], data: str) -> DC:
    """Instantiate a dataclass from a YAML string.

    Arguments:
        data_cls: A dataclass type.
        data: A YAML string.

    Returns:
        A `data_cls` instance.
    """
    return from_dict(data_cls, yaml.safe_load(data))


def to_yaml(data: Dataclass, **kwargs) -> str:  # noqa: ANN003
    """Serialize a dataclass instance to a YAML string.

    Arguments:
        data: A dataclass instance.
        kwargs: Keyword arguments passed to :func:`yaml.safe_dump`.

    Returns:
        A YAML string representation of `data`.
    """
    kwargs.setdefault("sort_keys", False)
    return yaml.safe_dump(to_dict(data, recursive=True), **kwargs)


def from_dotlist(data_cls: type[DC], data: list[str]) -> DC:
    """Instantiate a dataclass from a list of dot-style strings.

    Arguments:
        data_cls: A dataclass type.
        data: A list of dot-style strings.
            For example, `["foo.bar=1", "foo.bis=[baz,qux]"]`.

    Returns:
        A `data_cls` instance.
    """
    return from_omega(data_cls, OmegaConf.from_dotlist(data))



"""Heracls - Slayer of Hydra"""

__version__ = "0.5.1"

from . import cli, core, patch  # noqa: F401
from .cli import ArgumentParser, choice  # noqa: F401
from .core import (  # noqa: F401
    from_dict,
    from_dotlist,
    from_omega,
    from_yaml,
    to_dict,
    to_omega,
    to_yaml,
)
